import tempfile
import unittest
from pathlib import Path
import pyterrier as pt
import pyterrier_caching

class TestIndexerCache(unittest.TestCase):
    def setUp(self):
        if not pt.started():
            pt.init()

    def test_basic(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            cache = pyterrier_caching.IndexerCache(d/'cache')
            cache.index([
                {'docno': '1', 'data': 'test'},
                {'docno': '2', 'data': 'caching pyterrier'},
                {'docno': '3', 'data': 'fetch me some data'},
                {'docno': '4', 'data': 'information retrieval'},
                {'docno': '5', 'data': 'foo bar baz'},
            ])
            with self.subTest('full iter'):
                self.assertEqual(list(cache), [
                    {'docno': '1', 'data': 'test'},
                    {'docno': '2', 'data': 'caching pyterrier'},
                    {'docno': '3', 'data': 'fetch me some data'},
                    {'docno': '4', 'data': 'information retrieval'},
                    {'docno': '5', 'data': 'foo bar baz'},
                ])
            with self.subTest('start only'):
                self.assertEqual(list(cache.get_corpus_iter(start=2)), [
                    {'docno': '3', 'data': 'fetch me some data'},
                    {'docno': '4', 'data': 'information retrieval'},
                    {'docno': '5', 'data': 'foo bar baz'},
                ])
            with self.subTest('stop only'):
                self.assertEqual(list(cache.get_corpus_iter(stop=2)), [
                    {'docno': '1', 'data': 'test'},
                    {'docno': '2', 'data': 'caching pyterrier'},
                ])
            with self.subTest('start and stop'):
                self.assertEqual(list(cache.get_corpus_iter(start=1, stop=3)), [
                    {'docno': '2', 'data': 'caching pyterrier'},
                    {'docno': '3', 'data': 'fetch me some data'},
                ])
            with self.subTest('empty'):
                self.assertEqual(list(cache.get_corpus_iter(start=3, stop=1)), [])

    def test_empty(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            cache = pyterrier_caching.IndexerCache(d/'cache')
            cache.index(range(0))
            self.assertEqual(list(cache), [])



import pandas as pd
import numpy as np
import tempfile
import unittest
from pathlib import Path
import pyterrier as pt
import pyterrier_caching
from npids import Lookup

class TestScorerCache(unittest.TestCase):
    def setUp(self):
        if not pt.started():
            pt.init()

    def test_build_from_iter(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            cache = pyterrier_caching.ScorerCache(d/'cache')
            cache.build([
                {'docno': '1', 'data': 'test'},
                {'docno': '2', 'data': 'caching pyterrier'},
                {'docno': '3', 'data': 'fetch me some data'},
                {'docno': '4', 'data': 'information retrieval'},
                {'docno': '5', 'data': 'foo bar baz'},
            ])
            self.assertEqual(cache.corpus_count(), 5)

    def test_build_from_docnos(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            cache = pyterrier_caching.ScorerCache(d/'cache')
            docnos = Lookup.build(['1', '2', '3', '4', '5'], d/'docnos.npids')
            cache.build(docnos_file=d/'docnos.npids')
            self.assertEqual(cache.corpus_count(), 5)

    def test_basic(self):
        with tempfile.TemporaryDirectory() as d:
            d = Path(d)
            scorer_1 = pt.apply.score(lambda df: 1.)
            scorer_raises = pt.apply.score(lambda df: 0/0)
            scorer_2 = pt.apply.score(lambda df: 2.)
            cache = pyterrier_caching.ScorerCache(d/'cache', scorer_1)
            docnos = Lookup.build(['1', '2', '3', '4', '5'], d/'docnos.npids')
            cache.build(docnos_file=d/'docnos.npids')
            # seeding the cache:
            res = cache(pd.DataFrame([
                {'qid': 'a', 'query': 'a', 'docno': '1'},
                {'qid': 'a', 'query': 'a', 'docno': '2'},
                {'qid': 'b', 'query': 'b', 'docno': '2'},
            ]))
            self.assertTrue((res == pd.DataFrame([
                {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 0},
                {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0},
            ])).all().all())

            with self.subTest('make sure it doesn\'t call the scorer again on any of these'):
                cache.scorer = scorer_raises
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '2'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                ]))
                self.assertTrue((res == pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 0},
                    {'qid': 'a', 'query': 'a', 'docno': '2', 'score': 1., 'rank': 1},
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 0},
                ])).all().all())

            with self.subTest('make sure it calls a new scorer on any unknown values'):
                cache.scorer = scorer_2
                res = cache(pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1'},
                    {'qid': 'a', 'query': 'a', 'docno': '3'},
                    {'qid': 'b', 'query': 'b', 'docno': '2'},
                    {'qid': 'b', 'query': 'a', 'docno': '1'}, # should pull from (a,1) cache (not dependent on qid)
                    {'qid': 'c', 'query': 'c', 'docno': '1'},
                ]))
                self.assertTrue((res == pd.DataFrame([
                    {'qid': 'a', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 1}, # old
                    {'qid': 'a', 'query': 'a', 'docno': '3', 'score': 2., 'rank': 0}, # new
                    {'qid': 'b', 'query': 'a', 'docno': '1', 'score': 1., 'rank': 0}, # old
                    {'qid': 'b', 'query': 'b', 'docno': '2', 'score': 1., 'rank': 1}, # old
                    {'qid': 'c', 'query': 'c', 'docno': '1', 'score': 2., 'rank': 0}, # new
                ])).all().all())



from typing import Union
from pathlib import Path
from contextlib import contextmanager
import numpy as np
import pyterrier as pt

@contextmanager
def closing_memmap(*args, **kwargs):
    # np.memmap isn't allowed to be used as a context manager directly,
    # and the proper way to close it is by deleting the object.
    # This is a context manager that does that.
    mmp = None
    try:
        mmp = np.memmap(*args, **kwargs)
        yield mmp
    finally:
        if mmp is not None:
            del mmp
            mmp = None


class Lazy(pt.Transformer):
    def __init__(self, fn_transformer):
        self.fn_transformer = fn_transformer
        self.transformer = None

    def transform(self, inp):
        if not self.loaded():
            self.transformer = self.fn_transformer()
        return self.transformer(inp)

    def loaded(self):
        return self.transformer is not None

    def unload(self):
        self.transformer = None



from typing import Optional, List, Iterator, Dict, Any, Union
from pathlib import Path
from contextlib import ExitStack
import struct
import hashlib
import pickle
import lz4.frame
import numpy as np
import pandas as pd
import pyterrier as pt
import shutil
import json
from npids import Lookup
from pyterrier_caching import BuilderMode, artefact_builder


class Hdf5ScorerCache(pt.Transformer):
    artefact_type = 'scorer_cache'
    artefact_format = 'hdf5'

    def __init__(self, path, scorer=None):
        self.mode = 'r'
        self.path = Path(path)
        self.scorer = scorer
        self.meta = None
        self.file = None
        self.docnos = None
        self.dataset_cache = {}

    def transform(self, inp):
        self._ensure_built()
        results = []
        for query, group in inp.groupby('query'):
            query_hash = hashlib.sha256(query.encode()).hexdigest()
            ds = self._get_dataset(query_hash)
            dids = self.docnos.inv[np.array(group.docno)]
            dids_sorted, undo_did_sort = np.unique(dids, return_inverse=True)
            scores = ds[dids_sorted][undo_did_sort]
            to_score = group.loc[group.docno[np.isnan(scores)].index]
            if len(to_score) > 0:
                self._ensure_write_mode()
                ds = self._get_dataset(query_hash)
                new_scores = self.scorer(to_score)
                dids = self.docnos.inv[np.array(new_scores.docno)]
                dids_sorted, dids_sorted_idx = np.unique(dids, return_index=True)
                ds[dids_sorted] = new_scores.score.iloc[dids_sorted_idx]
                dids = self.docnos.inv[np.array(group.docno)]
                dids_sorted, undo_did_sort = np.unique(dids, return_inverse=True)
                scores = ds[dids_sorted][undo_did_sort]
            results.append(group.assign(score=scores))
        results = pd.concat(results, ignore_index=True)
        pt.model.add_ranks(results)
        return results

    def built(self) -> bool:
        return (Path(self.path)/'meta.json').exists()

    def build(self, corpus_iter=None, docnos_file=None):
        assert not self.built(), "this cache is alrady built"
        assert corpus_iter is not None or docnos_file is not None
        import h5py
        with artefact_builder(self.path, BuilderMode.create, self.artefact_type, self.artefact_format) as builder:
            with h5py.File(str(builder.path/'data.h5'), 'a'):
                pass # just create the data file
            if docnos_file:
                shutil.copy(docnos_file, builder.path/'docnos.npids')
                builder.metadata['doc_count'] = len(Lookup(builder.path/'docnos.npids'))
            else:
                builder.metadata['doc_count'] = 0
                with Lookup.builder(builder.path/'docnos.npids') as docno_lookup:
                    for record in corpus_iter:
                        docno_lookup.add(record['docno'])
                        builder.metadata['doc_count'] += 1

    def _ensure_built(self):
        import h5py
        assert self.built(), "you must .build(...) this cache before it can be used"
        if self.file is None:
            self.file = h5py.File(self.path/'data.h5', self.mode)
        if self.meta is None:
            with (self.path/'meta.json').open('rt') as fin:
                self.meta = json.load(fin)
        if self.docnos is None:
            self.docnos = Lookup(self.path/'docnos.npids')

    def _ensure_write_mode(self):
        if self.mode == 'r':
            assert self.scorer is not None, "missing value in cache; scorer is required"
            import h5py
            self.mode = 'a'
            self.file.close()
            self.file = h5py.File(self.path/'data.h5', self.mode)
            self.dataset_cache = {} # file changed, need to reset the cache

    def _get_dataset(self, qid):
        if qid not in self.dataset_cache:
            if qid not in self.file:
                self._ensure_write_mode()
                # TODO: setting chunks=(4096,) --- or some other value? --- might help
                # reduce the file size and/or speed up writes? Investigate more...
                self.file.create_dataset(qid, shape=(self.corpus_count(),), dtype=np.float32, fillvalue=float('nan'))
            self.dataset_cache[qid] = self.file[qid]
        return self.dataset_cache[qid]

    def corpus_count(self):
        self._ensure_built()
        return self.meta['doc_count']

# Default implementation of ScorerCache: Hdf5ScorerCache
ScorerCache = Hdf5ScorerCache



from typing import Optional, List, Iterator, Dict, Any, Union
from pathlib import Path
from contextlib import ExitStack
import struct
import pickle
import lz4.frame
import numpy as np
import pandas as pd
import pyterrier as pt
from npids import Lookup
from pyterrier_caching import BuilderMode, closing_memmap, artefact_builder


class Lz4PickleIndexerCache(pt.Indexer):
    artefact_type = 'indexer_cache'
    artefact_format = 'lz4pickle'

    def __init__(self, path: str):
        self.path = path

    def indexer(self, mode: Union[str, BuilderMode] = BuilderMode.create, skip_docno_lookup: bool = False) -> pt.Indexer:
        return Lz4PickleIndexerCacheIndexer(self, mode, skip_docno_lookup=skip_docno_lookup)

    def index(self, it: Iterator[Dict[str, Any]]) -> None:
        return self.indexer().index(it)

    def __iter__(self) -> Iterator[Dict[str, Any]]:
        return self.get_corpus_iter()

    def get_corpus_iter(self, verbose: bool = False, fields: Optional[List[str]] = None, start: Optional[int] = None, stop: Optional[int] = None) -> Iterator[Dict[str, Any]]:
        # validate arguments
        if fields is not None:
            fields = set(fields)
        if start is not None:
            assert start >= 0
        if stop is not None:
            assert stop >= 0
            # we take one extra one to get the offset of the last record, too.
            stop += 1

        with open(Path(self.path)/'data.pkl.lz4', mode='rb') as fdata, \
             closing_memmap(Path(self.path)/'offsets.np', dtype=np.uint64) as offsets_mmp:
            offsets = offsets_mmp[start:stop]
            if len(offsets) > 0:
                fdata.seek(offsets[0])
                it = zip(offsets[:-1], offsets[1:])
                if verbose:
                    it = tqdm(it, total=len(offsets)-1, unit='d')
                for offset_start, offset_stop in it:
                    record_length = offset_stop - offset_start
                    record = fdata.read(record_length)
                    record = lz4.frame.decompress(record)
                    record = pickle.loads(record)
                    if fields is not None:
                        record = {k: v for k, v in record.items() if k in fields}
                    yield record

    def to_dataframe(self, verbose: bool = False, fields: Optional[List[str]] = None, start: Optional[int] = None, stop: Optional[int] = None) -> pd.DataFrame:
        return pd.DataFrame(self.get_corpus_iter(verbose=verbose, fields=fields, start=start, stop=stop))

    def __getitem__(self, items: Union[int, str, slice]):
        if isinstance(items, int):
            return next(self.get_corpus_iter(start=items, stop=items+1))
        elif isinstance(items, str):
            idx = Lookup(Path(self.path)/'docnos.npids').inv[items]
            return self[idx]
            # TODO: support lookup by a list of strings?
        elif isinstance(items, slice):
            assert items.step is None, "step is not supported by Lz4PickleIndexerCache"
            return list(self.get_corpus_iter(start=items.start, stop=items.stop))
        raise ValueError('unknown type for items: {}'.format(type(items)))

    def built(self) -> bool:
        return (Path(self.path)/'meta.json').exists()


class Lz4PickleIndexerCacheIndexer(pt.Indexer):
    def __init__(self, cache: Lz4PickleIndexerCache, mode: Union[str, BuilderMode], skip_docno_lookup: bool = False):
        self.cache = cache
        self.mode = mode
        self.skip_docno_lookup = skip_docno_lookup

    def index(self, it: Iterator[Dict[str, Any]]) -> None:
        assert BuilderMode(self.mode) == BuilderMode.create, "Lz4PickleIndexerCache only supports 'create' mode"
        with ExitStack() as s:
            builder = s.enter_context(artefact_builder(self.cache.path, self.mode, self.cache.artefact_type, self.cache.artefact_format))
            fdata = s.enter_context(open(builder.path/'data.pkl.lz4', mode='wb'))
            foffsets = s.enter_context(open(builder.path/'offsets.np', 'wb'))
            docno_lookup = False if self.skip_docno_lookup else None
            foffsets.write(np.array(0, dtype=np.uint64).tobytes())
            builder.metadata['record_count'] = 0
            for record in it:
                if docno_lookup is None:
                    if 'docno' in record:
                        docno_lookup = s.enter_context(Lookup.builder(builder.path/'docnos.npids'))
                if docno_lookup is not False:
                    docno_lookup.add(record['docno'])
                record_bytes = pickle.dumps(dict(record))
                fdata.write(lz4.frame.compress(record_bytes))
                foffsets.write(np.array(fdata.tell(), dtype=np.uint64))
                builder.metadata['record_count'] += 1


# Default implementation of IndexerCache: Lz4PickleIndexerCache
IndexerCache = Lz4PickleIndexerCache



__version__ = '0.0.1'

from pyterrier_caching.util import Lazy, closing_memmap
from pyterrier_caching.builder import artefact_builder, BuilderMode
from pyterrier_caching.indexer_cache import IndexerCache, Lz4PickleIndexerCache
from pyterrier_caching.scorer_cache import ScorerCache, Hdf5ScorerCache



import setuptools


def get_version(path):
    import os
    for line in open(path, 'rt'):
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name='pyterrier-caching',
    version=get_version('pyterrier_caching/__init__.py'),
    author='Sean MacAvaney',
    author_email='sean.macavaney@glasgow.ac.uk',
    description='Caching components for PyTerrier',
    long_description=open('README.md', 'rt').read(),
    long_description_content_type='text/markdown',
    license='MIT',
    packages=setuptools.find_packages(),
    install_requires=list(open('requirements.txt', 'rt')),
    python_requires='>=3.6',
)



from typing import Union
import json
from enum import Enum
from pathlib import Path
from contextlib import contextmanager

class BuilderMode(Enum):
    create = 'create'
    overwrite = 'overwrite'
    append = 'append'


class ArtefactBuilderState:
    def __init__(self, path, metadata=None):
        self.path = Path(path)
        self.metadata = metadata or {}


@contextmanager
def artefact_builder(path: Union[str, Path], mode: Union[BuilderMode, str], artefact_type: str, artefact_format: str):
    mode = BuilderMode(mode)
    path = Path(path)

    state = ArtefactBuilderState(path, metadata={
        'type': artefact_type,
        'format': artefact_format,
    })

    # TODO: check if the path exists and either:
    if mode == BuilderMode.create:
        if path.exists():
            raise FileExistsError(f'{str(path)} already exists.')
        path.mkdir(parents=True)
    elif mode == BuilderMode.overwrite:
        if path.exists():
            raise NotImplementedError()
            # This might be a bit tricky... Do we recursively delete the entire
            # directory if it exists? What if the user accidently enters a path
            # such as '/'? Do we need some sort of check or warning if the path
            # is very large? Hmmm...
        else:
            path.mkdir(parents=True)
    elif mode == BuilderMode.append:
        raise NotImplementedError()

    try:
        yield state
    except:
        # TODO: clean up
        raise

    # Log the artefact metadata
    meta_path = path / 'meta.json'
    try:
        with open(meta_path, 'wt') as fout:
            json.dump(state.metadata, fout)
    except:
        if meta_path.exists():
            meta_path.unlink()
        raise
